#include "btv.h"

FILE *trace_file_ptr;
FILE *spd_file_ptr;

main(int argc, char *argv[]){

	FILE		*fout;
	STATS		stats;
	TRACE		trace;
	dram_system_t	dram_system;
	int		length,i,j;
	char 		c;
	int		analysisflag;
	char		filename[256];
	char		outfile[256];
	int		argc_index;
	int		t_burst;
	int		t_rc;
	int		min_t_rc;
	int		max_t_rc;
	int		t_rc_increment;
	int		t_faw;
	int		min_t_faw;
	int		max_t_faw;
	int		t_faw_increment;
	int		bank_count;
	int		min_bank_count;
	int		max_bank_count;
	int		bank_count_increment;
	int		queue_depth;
	int		min_queue_depth;
	int		max_queue_depth;
	int		queue_depth_increment;
	int		t_dqs;
	int		min_t_dqs;
	int		max_t_dqs;
	int		t_dqs_increment;
	int		rank_count;
	int		min_rank_count;
	int		max_rank_count;
	int		rank_count_increment;

	/* set defaults */
	
	argc_index 	= 1;
	analysisflag 	= PRINT_SUMMARY;
	trace.trace_type = UNKNOWN;
	trace_file_ptr	= NULL;
	spd_file_ptr 	= NULL;
	fout		= stdout;
	filename[0]	= '\0';
	t_burst		= 8;
	min_t_rc	= 32;
	max_t_rc	= 80;
	t_rc_increment	= 8;
	min_rank_count	= 1;
	max_rank_count	= 2;
	rank_count_increment = 1;
	min_bank_count	= 8;
	max_bank_count	= 16;
	bank_count_increment = 8;
	min_queue_depth	= 0;
	max_queue_depth = 0;
	queue_depth_increment = 2;
	min_t_dqs	= 0;
	max_t_dqs	= 6;
	t_dqs_increment	= 2;
	
	if((argc < 2) || (strncmp(argv[1], "-help",5) == 0)){
		fprintf(stdout,"Usage: %s -options optionswitch\n",argv[0]);
		fprintf(stdout,"-trace_type [k6|mase|tm]\n");
		fprintf(stdout,"-trace_file TRACE_FILENAME\n");
		fprintf(stdout,"-dram:spd_input SPD_FILENAME\n");
		fprintf(stdout,"-output_file OUTPUT_FILENAME\n");
		fprintf(stdout,"-analysis_type [PRINT_ALL|PRINT_SUMMARY|PRINT_DETAILS|DRAM_ANALYSIS]\n");
		fprintf(stdout,"-debug \n");
		_exit(0);
	}

	while(argc_index < argc){
		length = strlen(argv[argc_index]);
		if(strncmp(argv[argc_index], "-trace_type",length) == 0) {
			length = strlen(argv[argc_index+1]);
			if(strncmp(argv[argc_index+1], "mase",length) == 0) {
				trace.trace_type = MASE_TRACE;
			} else if(strncmp(argv[argc_index+1], "k6",length) == 0) {
				trace.trace_type = K6_TRACE;
			} else {
				fprintf(stdout,"Unknown trace type [%s]\n",argv[argc_index+1]);
			}
			argc_index += 2;
		} else if(strncmp(argv[argc_index], "-trace_file",length) == 0) {
			length = strlen(argv[argc_index+1]);
			if((trace_file_ptr=fopen(argv[argc_index+1],"r"))==NULL){
				printf("Error in opening tracefile %s\n",argv[argc_index+1]);
				_exit(3);
			}
			length = strlen(argv[argc_index+1]);
			for(i=0;i<length-4;i++){
				filename[i] = argv[argc_index+1][i];
			}
			filename[length-4] = '\0';
			argc_index += 2;
		} else if(strncmp(argv[argc_index], "-dram:spd_input",length) == 0) {
			length = strlen(argv[argc_index+1]);
			if((spd_file_ptr=fopen(argv[argc_index+1],"r"))==NULL){
				printf("Error in opening spd file %s\n",argv[argc_index+1]);
				_exit(3);
			}
			fprintf(stderr,"Not currently enabled.  Talk to me. David.\n");
			_exit(0);
			argc_index += 2;
		} else if(strncmp(argv[argc_index], "-output_file",length) == 0) {
			sprintf(filename,"%s",argv[argc_index+1]);
			argc_index += 2;
		} else if(strncmp(argv[argc_index], "-analysis_type",length) == 0) {
			if((strncmp(argv[argc_index+1], "printall", 8)== 0)) {
				analysisflag = PRINT_ALL;
			} else if((strncmp(argv[argc_index+1], "printsummary", 10)== 0)) {
				analysisflag = PRINT_SUMMARY;
			} else if((strncmp(argv[argc_index+1], "printdetails", 10)== 0)) {
				analysisflag = PRINT_DETAILS;
			} else if((strncmp(argv[argc_index+1], "dram", 4)== 0)) {
				analysisflag = DRAM_ANALYSIS;
			}
			argc_index += 2;
		}
	}

	if((trace.trace_type == K6_TRACE) && (read_k6_trace(trace_file_ptr,&(trace)) != BTV_OK)){
		printf("K6 Trace file open failed\n");
		_exit(0);
 	} else if((trace.trace_type == MASE_TRACE) && (read_mase_trace(trace_file_ptr,&(trace)) != BTV_OK)){
		printf("Mase Trace file open failed\n");
		_exit(0);
	}
	printf("size[%d] starttime[%lf] endtime[%lf]\n",
		trace.event_count,
		trace.start_time,
		trace.end_time);

	init_btv_stats(&(stats));
	stats.end_no = trace.event_count - 1;

	init_dram_system(&(dram_system));

	if(spd_file_ptr != NULL){
		read_dram_config_from_file(spd_file_ptr, &(dram_system.config));
	}

	if(analysisflag == PRINT_ALL){
		btv_gather_advanced_stats(&(stats), &(trace));
		sprintf(outfile,"%s.detail_stat",filename);
		if((fout=fopen(outfile,"w+"))==NULL){
			printf("Error in opening %s for output\n",filename);
			_exit(3);
		}
		btv_print_advanced_stats(&(stats), PRINT_DETAILS, fout);
		fclose(fout);
		sprintf(outfile,"%s.summary_stat",filename);
		if((fout=fopen(outfile,"w+"))==NULL){
			printf("Error in opening %s for output\n",filename);
			_exit(3);
		}
		btv_print_advanced_stats(&(stats), PRINT_SUMMARY, fout);
		fclose(fout);
	} else if(analysisflag == PRINT_DETAILS){
		btv_gather_advanced_stats(&(stats), &(trace));
		btv_print_advanced_stats(&(stats), analysisflag, fout);
		fclose(fout);
	} else if(analysisflag == PRINT_SUMMARY){
		btv_gather_advanced_stats(&(stats), &(trace));
		btv_print_advanced_stats(&(stats), analysisflag, fout);
		fclose(fout);
	} else if(analysisflag == DRAM_ANALYSIS){
		sprintf(outfile,"%s.txt",filename);
		if((fout=fopen(outfile,"w+"))==NULL){
			printf("Error in opening %s for output\n",outfile);
			_exit(3);
		}
		init_dram_system(&(dram_system));
		dram_system.config.row_buffer_management_policy	= OPEN_PAGE;
		dram_system.config.pa_mapping_policy  		= OPEN_PAGE_BASELINE;
		dram_system.config.t_burst			= t_burst;
		for(rank_count = min_rank_count; rank_count <= max_rank_count;rank_count+=rank_count_increment){
			dram_system.config.rank_count			= rank_count;
			for(bank_count = min_bank_count; bank_count <= max_bank_count;bank_count+=bank_count_increment){
				dram_system.config.bank_count			= bank_count;
				for(t_dqs = min_t_dqs; t_dqs <= max_t_dqs;t_dqs+=t_dqs_increment){
					dram_system.config.t_dqs	= t_dqs;
					dram_system.config.transaction_queue_depth = 0;
					for(t_rc = min_t_rc; t_rc <= max_t_rc; t_rc += t_rc_increment){
						dram_system.config.t_rc			= t_rc;
						dram_system.config.t_faw		= (int) (0.5 * t_rc);
						init_dram_stats(&(stats));
						reset_dram_system(&(dram_system));
						btv_dram_stats(&(trace),&(dram_system), &(stats));
						fprintf(fout,"%d ", dram_system.config.t_rc/dram_system.config.t_burst - 1);
						btv_print_dram_stats(&(stats), PRINT_BANDWIDTH, fout);
						fprintf(fout,"%d %2d %2d %2d\n", 
								dram_system.config.rank_count,
								dram_system.config.bank_count,
								4 * dram_system.config.t_rc / dram_system.config.t_faw,
								dram_system.config.t_dqs);
	
						dram_system.config.t_faw		= t_rc;
						init_dram_stats(&(stats));
						reset_dram_system(&(dram_system));
						btv_dram_stats(&(trace),&(dram_system), &(stats));
						fprintf(fout,"%d ", dram_system.config.t_rc/dram_system.config.t_burst - 1);
						btv_print_dram_stats(&(stats), PRINT_BANDWIDTH, fout);
						fprintf(fout,"%d %2d %2d %2d\n", 
								dram_system.config.rank_count,
								dram_system.config.bank_count,
								4 * dram_system.config.t_rc / dram_system.config.t_faw,
								dram_system.config.t_dqs);
					}
				}
			}
		}
		fclose(fout);
	}
}

